#![allow(dead_code)]

use std::time::Instant;
use itertools::Itertools;
use rayon::prelude::*;

#[derive(Clone)]
struct MapLine {
    dest_range_start: u64,
    source_range_start: u64,
    range_length: u64,
}

impl From<String> for MapLine {
    fn from(value: String) -> Self {
        let mut numbers = value
            .split(" ")
            .map(|number| number.parse::<u64>().unwrap());

        MapLine{
            dest_range_start: numbers.next().unwrap(),
            source_range_start: numbers.next().unwrap(),
            range_length: numbers.next().unwrap(),
        }
    }
}

impl MapLine {
    fn is_in_range(&self, number: u64) -> bool {
        return self.source_range_start <= number && self.source_range_start + self.range_length > number
    }

    fn convert(&self, number: u64) -> u64 {
        let diff = number - self.source_range_start;
        return self.dest_range_start + diff;
    }

    fn source_start(&self) -> u64 {
        self.source_range_start
    }

    fn source_end(&self) -> u64 {
        self.source_range_start + self.range_length
    }
}

#[derive(Clone)]
struct ConvertMap {
    from: String,
    to: String,

    lines: Vec<MapLine>
}

impl ConvertMap {
    fn new(key: String, data: Vec<&str>) -> Self {
        let convert_title = key.split_once("-to-").unwrap();

        ConvertMap{
            from: convert_title.0.to_owned(),
            to: convert_title.1.to_owned(),
            lines: data.iter().map(|line| MapLine::from(line.to_string())).collect()
        }
    }

    fn map(&self, data: u64) -> u64 {

        self.lines
            .iter()
            .filter(|convert| convert.is_in_range(data))
            .map(|convert| convert.convert(data))
            .next().unwrap_or(data)
    }

    fn from(&self) -> &str {
        &self.from
    }

    fn to(&self) -> &str {
        &self.to
    }
}


pub fn execute_task01(content: &str) {
    let start = Instant::now();
    let lowest_location = solve_01(content).unwrap();
    let duration = start.elapsed();

    assert_eq!(322500873, lowest_location);
    println!("Day05 - Task01 - Duration: {:?} - Lowest Location: {}", duration, lowest_location)
}

pub fn execute_task02(content: &str) {
    let start = Instant::now();
    let lowest_range_location = solve_02(content).unwrap();
    let duration = start.elapsed();

    assert_eq!(108956227, lowest_range_location);
    println!("Day05 - Task02 - Duration: {:?} - Lowest Seed Range Location: {}", duration, lowest_range_location)
}

pub fn solve_01(data: &str) -> anyhow::Result<u64> {
        let (seeds, convert_maps) = convert_input_to_convert_maps(data);

        Ok(seeds.iter()
            .map(|seed| convert_maps
                .iter()
                .fold(seed.clone(), |acc, map| map.map(acc)))
            .min().unwrap())
}

pub fn solve_02(data: &str) -> anyhow::Result<u64> {
    let (seeds, convert_maps) = convert_input_to_convert_maps(data);

    Ok(seeds
        .chunks(2)
        .into_iter()
        .map(|chunk| chunk.iter().collect_tuple().unwrap())
        .map(|(start, range)| *start..start+range)
        .par_bridge()
        .flat_map(|seed_range| {
            seed_range.clone().into_par_iter()
                .map(|seed| convert_maps
                    .iter()
                    .fold(seed, |acc, map| map.map(acc)))
        })
        .min().unwrap())
}

fn convert_input_to_convert_maps(input: &str) -> (Vec<u64>, Vec<ConvertMap>) {
    let mut lines = input.lines();
    let seed_line = lines.next().unwrap();
    lines.next();
    let mut maps: Vec<Vec<&str>> = vec![];

    let mut data = vec![];
    let mut key = String::new();

    for line in lines {
        if line.is_empty() {
            maps.push(data);
            key = String::new();
            data = vec![];
            continue;
        }

        if key.is_empty() {
            key = line.to_owned();
            continue;
        }

        data.push(line);
    }

    maps.push(data);

    let mut convert_map: Vec<ConvertMap> = vec![];
    for map in maps {
        let map = ConvertMap::new("data-to-data".to_owned(), map);
        convert_map.push(map);
    }


    (seed_line
        .split_once(": ").unwrap().1
        .split(" ")
         .map(|data| data.trim())
         .filter(|data| data.len() > 0)
        .map(|num| num.parse::<u64>().unwrap())
        .collect(),
     convert_map)
}


#[test]
pub fn test_solve_01() {
    let example = r#"seeds: 79 14 55 13

seed-to-soil map:
50 98 2
52 50 48

soil-to-fertilizer map:
0 15 37
37 52 2
39 0 15

fertilizer-to-water map:
49 53 8
0 11 42
42 0 7
57 7 4

water-to-light map:
88 18 7
18 25 70

light-to-temperature map:
45 77 23
81 45 19
68 64 13

temperature-to-humidity map:
0 69 1
1 0 69

humidity-to-location map:
60 56 37
56 93 4"#;

    let result = solve_01(example).unwrap();

    assert_eq!(35, result);
}

#[test]
pub fn test_solve_02() {
    let example = r#"seeds: 79 14 55 13

seed-to-soil map:
50 98 2
52 50 48

soil-to-fertilizer map:
0 15 37
37 52 2
39 0 15

fertilizer-to-water map:
49 53 8
0 11 42
42 0 7
57 7 4

water-to-light map:
88 18 7
18 25 70

light-to-temperature map:
45 77 23
81 45 19
68 64 13

temperature-to-humidity map:
0 69 1
1 0 69

humidity-to-location map:
60 56 37
56 93 4"#;

    let result = solve_02(example).unwrap();

    assert_eq!(46, result);
}